import unittest
import numpy as np
from dirichletcal.calib.fixeddirichlet import FixedDiagonalDirichletCalibrator
from . import get_simple_binary_example
from . import get_simple_ternary_example


class TestFixedDiagonalDirichlet(unittest.TestCase):
    def setUp(self):
        self.cal = FixedDiagonalDirichletCalibrator()

    #def test_fit_predict(self):
    #    S, y = get_simple_binary_example()
    #    self.cal.fit(S, y)
    #    predictions = self.cal.predict_proba(S).argmax(axis=1)
    #    np.testing.assert_array_equal(predictions, y)

    #    S, y = get_simple_ternary_example()
    #    self.cal.fit(S, y)
    #    predictions = self.cal.predict_proba(S).argmax(axis=1)
    #    np.testing.assert_array_equal(predictions, y)


def main():
    unittest.main()


if __name__ == '__main__':
    main()



from .version import __version__
import numpy as np

from sklearn.base import BaseEstimator, RegressorMixin

from .calib.fulldirichlet import FullDirichletCalibrator
from .calib.diagdirichlet import DiagonalDirichletCalibrator
from .calib.fixeddirichlet import FixedDiagonalDirichletCalibrator


class DirichletCalibrator(BaseEstimator, RegressorMixin):
    def __init__(self, matrix_type='full', l2=0.0, comp_l2=False,
                 initializer='identity'):
        if matrix_type not in ['full', 'full_gen', 'diagonal',
                               'fixed_diagonal']:
            raise ValueError
        self.matrix_type = matrix_type
        self.l2 = l2
        self.comp_l2 = comp_l2
        self.initializer = initializer


    def __setup(self):
        if isinstance(self.l2, list):
            self.l2_grid = self.l2
        else:
            self.l2_grid = [self.l2]
        if isinstance(self.comp_l2, list):
            self.comp_l2 = self.comp_l2
        else:
            self.comp_l2 = [self.comp_l2]
        self.calibrator_ = None


    def fit(self, x, y, x_val=None, y_val=None, **kwargs):
        self.__setup()

        if self.matrix_type == 'diagonal':
            self.calibrator_ = DiagonalDirichletCalibrator(
                l2=self.l2, initializer=self.initializer)
        elif self.matrix_type == 'fixed_diagonal':
            self.calibrator_ = FixedDiagonalDirichletCalibrator(
                l2=self.l2, initializer=self.initializer)
        elif self.matrix_type == 'full':
            self.calibrator_ = FullDirichletCalibrator(
                reg_lambda_list=self.l2_grid, reg_mu_list=self.comp_l2,
                initializer=self.initializer)
        elif self.matrix_type == 'full_gen':
            self.calibrator_ = GenerativeDirichletCalibrator()
        else:
            raise ValueError

        _X = np.copy(x)
        if len(x.shape) == 1:
            _X = np.vstack(((1 - _X), _X)).T

        _X_val = x_val
        if x_val is not None:
            _X_val = np.copy(x_val)
            if len(x_val.shape) == 1:
                _X_val = np.vstack(((1 - _X_val), _X_val)).T

        self.calibrator_ = self.calibrator_.fit(_X, y, X_val=_X_val,
                                                y_val=y_val, **kwargs)
        return self

    @property
    def l2_(self):
        if (self.calibrator_ is not None) and (hasattr(self.calibrator_, 'l2')):
            return self.calibrator_.l2
        return None

    @property
    def weights_(self):
        if (self.calibrator_ is not None) and (hasattr(self.calibrator_, 'weights_')):
            return self.calibrator_.weights_
        return None

    @property
    def coef_(self):
        if (self.calibrator_ is not None) and (hasattr(self.calibrator_, 'coef_')):
            return self.calibrator_.coef_
        return None

    @property
    def intercept_(self):
        if (self.calibrator_ is not None) and (hasattr(self.calibrator_, 'intercept_')):
            return self.calibrator_.intercept_
        return None

    @property
    def cannonical_weights(self):
        b = self.weights_[:, -1]
        w = self.weights_[:, :-1]
        col_min = np.min(w, axis=0)
        a = w - col_min

        def softmax(z):
            return np.divide(np.exp(z), np.sum(np.exp(z)))

        c = softmax(np.matmul(w, np.log(np.ones(len(b)) / len(b))) + b)
        return np.hstack((a, c.reshape(-1, 1)))

    def predict_proba(self, s):

        _s = np.copy(s)
        if len(s.shape) == 1:
            _s = np.vstack(((1 - _s), _s)).T
            return self.calibrator_.predict_proba(_s)[:, 1]

        return self.calibrator_.predict_proba(_s)

    def predict(self, s):

        _s = np.copy(s)
        if len(s.shape) == 1:
            _s = np.vstack(((1 - _s), _s)).T
            return self.calibrator_.predict(_s)[:, 1]

        return self.calibrator_.predict(_s)



from __future__ import division

import logging

import jax.numpy as np
import jax
import numpy as raw_np
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.preprocessing import label_binarize

import scipy
import scipy.optimize
import scipy.linalg

from ..utils import clip_jax

from jax.config import config
config.update("jax_enable_x64", True)


class MultinomialRegression(BaseEstimator, RegressorMixin):
    def __init__(self, weights_0=None, method='Full', initializer='identity',
                 reg_format=None, reg_lambda=0.0, reg_mu=None, reg_norm=False,
                 ref_row=True, optimizer='auto'):
        """
        Params:
            optimizer: string ('auto', 'newton', 'fmin_l_bfgs_b')
                If 'auto': then 'newton' for less than 37 classes and
                fmin_l_bfgs_b otherwise
                If 'newton' then uses our implementation of a Newton method
                If 'fmin_l_bfgs_b' then uses scipy.ptimize.fmin_l_bfgs_b which
                implements a quasi Newton method
        """
        if method not in ['Full', 'Diag', 'FixDiag']:
            raise(ValueError('method {} not avaliable'.format(method)))

        self.weights_0 = weights_0
        self.method = method
        self.initializer = initializer
        self.reg_format = reg_format
        self.reg_lambda = reg_lambda
        self.reg_mu = reg_mu  # If number, then ODIR is applied
        self.reg_norm = reg_norm
        self.ref_row = ref_row
        self.optimizer = optimizer

    def __setup(self):
        self.classes = None
        self.weights_ = self.weights_0
        self.weights_0_ = self.weights_0

    @property
    def coef_(self):
        return self.weights_[:, :-1]

    @property
    def intercept_(self):
        return self.weights_[:, -1]

    def predict_proba(self, S):

        S_ = np.hstack((S, np.ones((len(S), 1))))

        return np.asarray(_calculate_outputs(self.weights_, S_))

    # FIXME Should we change predict for the argmax?
    def predict(self, S):

        return np.asarray(self.predict_proba(S))


    def fit(self, X, y, *args, **kwargs):

        self.__setup()

        X_ = np.hstack((X, np.ones((len(X), 1))))

        self.classes = raw_np.unique(y)

        k = len(self.classes)

        if self.reg_norm:
            if self.reg_mu is None:
                self.reg_lambda = self.reg_lambda / (k * (k + 1))
            else:
                self.reg_lambda = self.reg_lambda / (k * (k - 1))
                self.reg_mu = self.reg_mu / k

        target = label_binarize(y, self.classes)

        if k == 2:
            target = np.hstack([1-target, target])

        n, m = X_.shape

        XXT = (X_.repeat(m, axis=1) * np.hstack([X_]*m)).reshape((n, m, m))

        logging.debug(self.method)

        self.weights_0_ = self._get_initial_weights(self.initializer)

        if (self.optimizer == 'newton'
            or (self.optimizer == 'auto' and k <= 36)):
            weights = _newton_update(self.weights_0_, X_, XXT, target, k,
                                     self.method, reg_lambda=self.reg_lambda,
                                     reg_mu=self.reg_mu, ref_row=self.ref_row,
                                     initializer=self.initializer,
                                     reg_format=self.reg_format)
        elif (self.optimizer == 'fmin_l_bfgs_b'
              or (self.optimizer == 'auto' and k > 36)):

            _gradient_np = lambda *args, **kwargs: raw_np.array(_gradient(*args, **kwargs))

            res = scipy.optimize.fmin_l_bfgs_b(func=_objective,
                                               fprime=_gradient_np,
                                               x0=self.weights_0_,
                                               args=(X_, XXT, target, k,
                                                     self.method,
                                                     self.reg_lambda,
                                                     self.reg_mu, self.ref_row,
                                                     self.initializer,
                                                     self.reg_format),
                                               maxls=128,
                                               factr=1.0)
            weights = res[0]
        else:
            raise(ValueError('Unknown optimizer: {}'.format(self.optimizer)))

        self.weights_ = _get_weights(weights, k, self.ref_row, self.method)

        return self

    def _get_initial_weights(self, ref_row, initializer='identity'):
        ''' Returns an array containing only the weights of the full weight
        matrix.

        '''

        if initializer not in ['identity', None]:
            raise ValueError

        k = len(self.classes)

        if self.weights_0_ is None:
            if initializer == 'identity':
                weights_0 = _get_identity_weights(k, ref_row, self.method)
            else:
                if self.method == 'Full':
                    weights_0 = np.zeros(k * (k + 1))
                elif self.method == 'Diag':
                    weights_0 = np.zeros(2*k)
                elif self.method == 'FixDiag':
                    weights_0 = np.zeros(1)
        else:
            weights_0 = self.weights_0_

        return weights_0


def _objective(params, *args):
    (X, _, y, k, method, reg_lambda, reg_mu, ref_row, _, reg_format) = args
    weights = _get_weights(params, k, ref_row, method)
    outputs = clip_jax(_calculate_outputs(weights, X))
    loss = np.mean(-np.log(np.sum(y * outputs, axis=1)))

    if reg_mu is None:
        if reg_format == 'identity':
            reg = np.hstack([np.eye(k), np.zeros((k, 1))])
        else:
            reg = np.zeros((k, k+1))
        loss = loss + reg_lambda * np.sum((weights - reg)**2)
    else:
        weights_hat = weights - np.hstack([weights[:, :-1] * np.eye(k),
                                           np.zeros((k, 1))])
        loss = loss + reg_lambda * np.sum(weights_hat[:, :-1] ** 2) + \
            reg_mu * np.sum(weights_hat[:, -1] ** 2)

    return loss


_gradient = jax.grad(_objective, argnums=0)


_hessian = jax.hessian(_objective, argnums=0)


def _get_weights(params, k, ref_row, method):
        ''' Reshapes the given params (weights) into the full matrix including 0
        '''

        if method in ['Full', None]:
            raw_weights = params.reshape(-1, k+1)
            # weights = np.zeros([k, k+1])
            # weights[:-1, :] = params.reshape(-1, k + 1)

        elif method == 'Diag':
            raw_weights = np.hstack([np.diag(params[:k]),
                                     params[k:].reshape(-1, 1)])
            # weights[:, :-1][np.diag_indices(k)] = params[:]

        elif method == 'FixDiag':
            raw_weights = np.hstack([np.eye(k) * params[0], np.zeros((k, 1))])
            # weights[np.dgag_indices(k - 1)] = params[0]
            # weights[np.diag_indices(k)] = params[0]
        else:
            raise(ValueError("Unknown calibration method {}".format(method)))

        if ref_row:
            weights = raw_weights - np.repeat(
                raw_weights[-1, :].reshape(1, -1), k, axis=0)
        else:
            weights = raw_weights

        return weights


def _get_identity_weights(n_classes, ref_row, method):

    raw_weights = None

    if (method is None) or (method == 'Full'):
        raw_weights = np.zeros((n_classes, n_classes + 1)) + \
                      np.hstack([np.eye(n_classes), np.zeros((n_classes, 1))])
        raw_weights = raw_weights.ravel()

    elif method == 'Diag':
        raw_weights = np.hstack([np.ones(n_classes), np.zeros(n_classes)])

    elif method == 'FixDiag':
        raw_weights = np.ones(1)

    return raw_weights.ravel()


def _calculate_outputs(weights, X):
    mul = np.dot(X, weights.transpose())
    return _softmax(mul)


def _softmax(X):
    """Compute the softmax of matrix X in a numerically stable way."""
    shiftx = X - np.max(X, axis=1).reshape(-1, 1)
    exps = np.exp(shiftx)
    return exps / np.sum(exps, axis=1).reshape(-1, 1)


def _newton_update(weights_0, X, XX_T, target, k, method_, maxiter=int(1024),
                   ftol=1e-12, gtol=1e-8, reg_lambda=0.0, reg_mu=None,
                   ref_row=True, initializer=None, reg_format=None):

    L_list = [float(_objective(weights_0, X, XX_T, target, k, method_,
                                      reg_lambda, reg_mu, ref_row, initializer,
                                      reg_format))]

    weights = weights_0.copy()

    # TODO move this to the initialization
    if method_ is None:
        weights = np.zeros_like(weights)

    for i in range(0, maxiter):

        gradient = _gradient(weights, X, XX_T, target, k, method_, reg_lambda,
                             reg_mu, ref_row, initializer, reg_format)

        if np.abs(gradient).sum() < gtol:
            break

        # FIXME hessian is ocasionally NaN
        hessian = _hessian(weights, X, XX_T, target, k, method_, reg_lambda,
                           reg_mu, ref_row, initializer, reg_format)

        if method_ == 'FixDiag':
            updates = gradient / hessian
        else:
            try:
                inverse = scipy.linalg.pinv2(hessian)
                updates = np.matmul(inverse, gradient)
            except (raw_np.linalg.LinAlgError, ValueError) as err:
                logging.error(err)
                updates = gradient

        for step_size in np.hstack((np.linspace(1, 0.1, 10),
                                    np.logspace(-2, -32, 31))):

            tmp_w = weights - (updates * step_size).ravel()

            if np.any(np.isnan(tmp_w)):
                logging.debug("{}: There are NaNs in tmp_w".format(method_))

            L = _objective(tmp_w, X, XX_T, target, k, method_, reg_lambda,
                           reg_mu, ref_row, initializer, reg_format)

            if (L - L_list[-1]) < 0:
                break

        L_list.append(float(L))

        logging.debug("{}: after {} iterations log-loss = {:.7e}, sum_grad = {:.7e}".format(
            method_, i, L, np.abs(gradient).sum()))

        if np.isnan(L):
            logging.error("{}: log-loss is NaN".format(method_))
            break

        if i >= 5:
            if (float(raw_np.min(raw_np.diff(L_list[-5:]))) > -ftol) & \
               (float(raw_np.sum(raw_np.diff(L_list[-5:])) > 0) == 0):
                weights = tmp_w.copy()
                logging.debug('{}: Terminate as there is not enough changes on loss.'.format(
                    method_))
                break

        if (L_list[-1] - L_list[-2]) > 0:
            logging.debug('{}: Terminate as the loss increased {}.'.format(
                method_, np.diff(L_list[-2:])))
            break
        else:
            weights = tmp_w.copy()

    L = _objective(weights, X, XX_T, target, k, method_,
                   reg_lambda, reg_mu, ref_row, initializer, reg_format)

    logging.debug("{}: after {} iterations final log-loss = {:.7e}, sum_grad = {:.7e}".format(
        method_, i, L, np.abs(gradient).sum()))

    return weights



import unittest
import numpy as np
from dirichletcal.calib.multinomial import MultinomialRegression
from dirichletcal.calib.multinomial import _get_weights
from . import get_simple_binary_example
from . import get_simple_ternary_example

from sklearn.datasets import make_classification
from sklearn.metrics import accuracy_score


class TestMultinomial(unittest.TestCase):
    def test_fit_predict(self):
        S, y = make_classification(n_samples=1000, n_classes=2, n_features=2,
                                   n_informative=2, n_redundant=0,
                                   n_clusters_per_class=1, scale=100,
                                   class_sep=100.0, random_state=42)

        mlr = MultinomialRegression()
        mlr.fit(S, y)
        predictions = mlr.predict_proba(S).argmax(axis=1)
        acc = accuracy_score(y, predictions)
        self.assertGreater(acc, 0.99, "accuracy must be superior to 99 percent")

        S, y = make_classification(n_samples=1000, n_classes=3, n_features=3,
                                   n_informative=3, n_redundant=0,
                                   n_clusters_per_class=1, scale=100,
                                   class_sep=100.0, random_state=243)
        mlr = MultinomialRegression()
        mlr.fit(S, y)
        predictions = mlr.predict_proba(S).argmax(axis=1)
        acc = accuracy_score(y, predictions)
        self.assertGreater(acc, 0.9, "accuracy must be superior to 9 percent")

    def test_get_weights_no_ref_row(self):
        k = 3
        params = np.arange(k*(k+1)) + 1
        full_matrix = _get_weights(params, k=k, ref_row=False, method='Full')
        expected = params.reshape(k, k+1)
        np.testing.assert_array_equal(full_matrix, expected)

        k = 3
        params = np.arange(k) + 1
        full_matrix = _get_weights(params, k=k, ref_row=False, method='FixDiag')
        expected = np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0]], 'float')
        np.testing.assert_array_equal(full_matrix, expected)

    def test_get_weights_ref_row(self):
        k = 5
        params = np.arange(k*(k+1)) + 1
        full_matrix = _get_weights(params, k=k, ref_row=True, method='Full')
        expected = params.reshape(k, k+1)
        expected -= expected[-1, :]
        np.testing.assert_array_equal(full_matrix, expected)

        k = 3
        params = np.arange(k) + 1
        full_matrix = _get_weights(params, k=k, ref_row=True, method='FixDiag')
        expected = np.hstack((np.eye(k), np.zeros((k, 1))))
        expected -= expected[-1, :]
        np.testing.assert_array_equal(full_matrix, expected)

    def test_optimizers(self):
        S, y = make_classification(n_samples=20, n_classes=2, n_features=2,
                                   n_informative=2, n_redundant=0,
                                   n_clusters_per_class=1, scale=10,
                                   class_sep=100.0, random_state=42)

        mlr = MultinomialRegression(optimizer='newton')
        mlr.fit(S, y)
        # TODO: add some test here

        mlr = MultinomialRegression(optimizer='fmin_l_bfgs_b')
        mlr.fit(S, y)
        # TODO: add some test here


def main():
    unittest.main()


if __name__ == '__main__':
    main()



import numpy as np

from .multinomial import MultinomialRegression
from .fulldirichlet import FullDirichletCalibrator
from ..utils import clip_for_log


class DiagonalDirichletCalibrator(FullDirichletCalibrator):

     def fit(self, X, y, *args, **kwargs):
        X_ = np.log(clip_for_log(X))
        self.calibrator_ = MultinomialRegression(method='Diag', l2=self.l2).fit(X_, y, *args, **kwargs)
        return self



import logging
from sklearn.base import BaseEstimator, RegressorMixin

import numpy as np
from .multinomial import MultinomialRegression
from ..utils import clip_for_log
from sklearn.metrics import log_loss

from .multinomial import _get_identity_weights


class FullDirichletCalibrator(BaseEstimator, RegressorMixin):
    def __init__(self, reg_lambda=0.0, reg_mu=None, weights_init=None,
                 initializer='identity', reg_norm=False, ref_row=True,
                 optimizer='auto'):

        """
        Params:
            weights_init: (nd.array) weights used for initialisation, if None
            then idendity matrix used. Shape = (n_classes - 1, n_classes + 1)
            comp_l2: (bool) If true, then complementary L2 regularization used
            (off-diagonal regularization)
            optimizer: string ('auto', 'newton', 'fmin_l_bfgs_b')
                If 'auto': then 'newton' for less than 37 classes and
                fmin_l_bfgs_b otherwise
                If 'newton' then uses our implementation of a Newton method
                If 'fmin_l_bfgs_b' then uses scipy.ptimize.fmin_l_bfgs_b which
                implements a quasi Newton method
        """
        self.reg_lambda = reg_lambda
        self.reg_mu = reg_mu  # Complementary L2 regularization. (Off-diagonal regularization)
        self.weights_init = weights_init  # Input weights for initialisation
        self.initializer = initializer
        self.reg_norm = reg_norm
        self.ref_row = ref_row
        self.optimizer = optimizer

    def fit(self, X, y, X_val=None, y_val=None, *args, **kwargs):

        self.weights_ = self.weights_init

        k = np.shape(X)[1]

        if X_val is None:
            X_val = X.copy()
            y_val = y.copy()

        _X = np.copy(X)
        _X = np.log(clip_for_log(_X))
        _X_val = np.copy(X_val)
        _X_val = np.log(clip_for_log(X_val))

        self.calibrator_ = MultinomialRegression(method='Full',
                                        reg_lambda=self.reg_lambda,
                                        reg_mu=self.reg_mu,
                                        reg_norm=self.reg_norm,
                                        ref_row=self.ref_row,
                                        optimizer=self.optimizer)
        self.calibrator_.fit(_X, y, *args, **kwargs)
        final_loss = log_loss(y_val, self.calibrator_.predict_proba(_X_val))

        return self

    @property
    def weights(self):
        if self.calibrator_ is not None:
            return self.calibrator_.weights_
        return self.weights_init

    @property
    def coef_(self):
        return self.calibrator_.coef_

    @property
    def intercept_(self):
        return self.calibrator_.intercept_

    def predict_proba(self, S):
        S = np.log(clip_for_log(S))
        return np.asarray(self.calibrator_.predict_proba(S))

    def predict(self, S):
        S = np.log(clip_for_log(S))
        return np.asarray(self.calibrator_.predict(S))



import numpy as np

import jax.numpy as jax_np

def clip_for_log(X):
    eps = np.finfo(X.dtype).tiny
    return np.clip(X, eps, 1-eps)


def clip(X):
    eps = np.finfo(X.dtype).tiny
    return np.clip(X, eps, 1-eps)

def clip_jax(X):
    eps = jax_np.finfo(X.dtype).eps
    return jax_np.clip(X, eps, 1-eps)



import unittest
from dirichletcal.calib.fulldirichlet import FullDirichletCalibrator
from . import get_simple_binary_example
from . import get_extreme_binary_example
from . import get_simple_ternary_example

from sklearn.metrics import accuracy_score


class TestFullDirichlet(unittest.TestCase):
    def setUp(self):
        self.cal = FullDirichletCalibrator()

    def test_fit_predict(self):
        for S, y in (get_simple_binary_example(),
                     get_simple_ternary_example()):
            self.cal = FullDirichletCalibrator()
            self.cal.fit(S, y)
            predictions = self.cal.predict_proba(S).argmax(axis=1)
            acc = accuracy_score(y, predictions)
            self.assertGreater(acc, 0.97,
                               "accuracy must be superior to 97 percent")

    def test_extreme_values(self):
        S, y = get_extreme_binary_example()
        self.cal = FullDirichletCalibrator()
        self.cal.fit(S, y)
        predictions = self.cal.predict_proba(S).argmax(axis=1)
        acc = accuracy_score(y, predictions)
        self.assertGreater(acc, 0.99,
                           "accuracy must be superior to 99 percent")

    def test_lambda(self):
        S, y = get_simple_ternary_example()
        l2_odir = 1e-2
        cal = FullDirichletCalibrator(reg_lambda=l2_odir,
                                      reg_mu=l2_odir, reg_norm=False)
        cal.fit(S, y)
        predictions = cal.predict_proba(S).argmax(axis=1)
        acc = accuracy_score(y, predictions)
        self.assertGreater(acc, 0.98,
                           "accuracy must be superior to 99 percent")

        l2_odir = 1e2
        cal = FullDirichletCalibrator(reg_lambda=l2_odir,
                                      reg_mu=l2_odir, reg_norm=False)
        cal.fit(S, y)
        predictions = cal.predict_proba(S).argmax(axis=1)
        acc = accuracy_score(y, predictions)
        self.assertLess(acc, 0.6,
                        "accuracy must be smaller than 60 percent")


def main():
    unittest.main()


if __name__ == '__main__':
    main()



import numpy as np
from sklearn.datasets import make_classification

from math import gamma
from operator import mul
import numpy as np
from functools import reduce

np.random.seed(42)

class Dirichlet(object):
    '''
    Based on http://blog.bogatron.net/blog/2014/02/02/visualizing-dirichlet-distributions/
    '''
    def __init__(self, alpha):
        self._alpha = np.array(alpha)
        self._coef = gamma(np.sum(self._alpha)) / \
                     reduce(mul, [gamma(a) for a in self._alpha])
    def pdf(self, x):
        '''Returns pdf value for `x`.'''
        return self._coef * reduce(mul, [xx ** (aa - 1)
                                         for (xx, aa)in zip(x, self._alpha)])

    def sample(self, size=None, **kwargs):
        return np.random.dirichlet(self._alpha, size=size, **kwargs)

    def __str__(self):
        return np.array2string(self._alpha, separator=',', precision=2)

    def __repr__(self):
        return self._alpha.__repr__()


class MixtureDistribution(object):
    def __init__(self, priors, distributions):
        self.priors = np.array(priors)
        self.distributions = distributions

    def sample(self, size=None):
        if size is None:
            size = len(self.priors)
        classes = np.random.multinomial(n=1, pvals=self.priors, size=size)
        samples = np.empty_like(classes, dtype='float')
        for i, size in enumerate(classes.sum(axis=0)):
            samples[np.where(classes[:,i])[0]] = self.distributions[i].sample(size)
        return samples, classes

    def posterior(self, pvalues, c=0):
        likelihoods = np.array([d.pdf(pvalues) for d in self.distributions])
        Z = np.dot(likelihoods, self.priors)
        return np.divide(likelihoods[c]*self.priors[c], Z)

    def pdf(self, pvalues):
        likelihoods = np.array([d.pdf(pvalues) for d in self.distributions])
        return np.dot(self.priors, likelihoods)

    def __repr__(self):
        string = ''
        for p, d in zip(self.priors, self.distributions):
            string += 'prior = {}, '.format(p)
            string += 'Distribution = {}'.format(d)
            string += '\n'
        return string

def get_simple_binary_example():
    S = np.random.beta(10, 0.1, 100).transpose()
    S[50:] = 1-S[50:]
    S = np.vstack([S, 1-S]).T
    y = np.hstack([np.zeros(50), np.ones(50)])
    return S, y

def get_extreme_binary_example():
    S = np.random.beta(10000, 0.1, 100).transpose()
    S[50:] = 1-S[50:]
    S = np.vstack([S, 1-S]).T
    y = np.hstack([np.zeros(50), np.ones(50)])
    return S, y

def get_simple_ternary_example():
    mdir = MixtureDistribution([1./3, 1./3, 1./3],
                               [Dirichlet([5, 0.2, .3]),
                                Dirichlet([1, 5, .5]),
                                Dirichlet([.1, 2, 10])])
    S, Y = mdir.sample(1000)
    y = Y.argmax(axis=1)
    return S, y



import autograd
import autograd.numpy as np

from sklearn.base import BaseEstimator, RegressorMixin

def clip_for_log(X):
    eps = np.finfo(X.dtype).tiny
    return np.clip(X, eps, 1-eps)

class FixedDiagonalDirichletCalibrator(BaseEstimator, RegressorMixin):
    def fit(self, X, y, batch_size=128, lr=1e-3, beta_1=0.9, beta_2=0.999,
            eeps=1e-8, maxiter=int(1024), factor=1e-4, *args, **kwargs):
        eps = np.finfo(X.dtype).min
        X_ = np.log(clip_for_log(X))
        k = np.shape(X_)[1]
        n_y = np.shape(X_)[0]
        y_hot = np.zeros((len(y), k))
        for i in range(0, k):
            y_hot[y==i, i] = 1.0
        L = []
        m = np.zeros(1)
        v = np.zeros(1)
        T = 1.0
        fin_T = T
        if batch_size is None:
            batch_size = n_y

        get_gradient = autograd.grad(self._objective, 0)

        batch_idx = np.hstack([np.arange(0, n_y, batch_size), n_y])

        batch_num = len(batch_idx) - 1

        for i in range(0, maxiter):

            batch_L = []

            for j in range(0, batch_num):

                L_t = self._objective(T, X, y_hot)

                g_t = get_gradient(T, X, y_hot)

                m = beta_1 * m + (1 - beta_1) * g_t

                v = beta_2 * v + (1 - beta_2) * g_t * g_t

                T = T - lr * m / (v ** 0.5 + eeps)

                batch_L.append(L_t)

                # print('Iteration: ' + str(i) + ', Batch: ' + str(j) + ', Theta: ' + str(theta[:8]))

            L.append(np.sum(batch_L))

            per_idx = np.random.permutation(n_y)

            X = X[per_idx]

            y_hot = y_hot[per_idx]

            y = y[per_idx]

            if len(L) >= 2:
                if L[-1] < np.min(L[:-1]):
                    fin_T = T.copy()

            if len(L) > 32:

                previous_opt = np.min(L.copy()[:-32])

                current_opt = np.min(L.copy()[-32:])

                if previous_opt - current_opt <= np.abs(previous_opt * factor):
                    break

        self.weights_ = fin_T

        return self

    def _objective(self, T, X, y_hot):

        tmp_prod = T * np.log(X)

        prob_y = np.exp(tmp_prod) / np.sum(np.exp(tmp_prod), axis=1).reshape(-1, 1)

        loss = -np.sum(np.log(np.sum(prob_y * y_hot, axis=1)))

        return loss

    @property
    def coef_(self):
        return self.weights_

    @property
    def intercept_(self):
        return self.calibrator_.intercept_

    def predict_proba(self, S):
        S = np.log(clip_for_log(S))
        k = np.shape(S)[1]
        tmp_prod = self.weights_ * S
        prob_y = np.exp(tmp_prod) / np.sum(np.exp(tmp_prod), axis=1).reshape(-1, 1)
        return prob_y

    def predict(self, S):
        S = np.log(clip_for_log(S))
        k = np.shape(S)[1]
        tmp_prod = self.weights_ * S
        prob_y = np.exp(tmp_prod) / np.sum(np.exp(tmp_prod), axis=1).reshape(-1, 1)
        return prob_y


if __name__ == '__main__':
    from sklearn import datasets

    # import some data to play with
    iris = datasets.load_iris()
    X = iris.data[:, :3]  # we only take the first two features.
    y = iris.target

    softmax = lambda z:np.divide(np.exp(z).T, np.sum(np.exp(z), axis=1)).T
    S = softmax(X)

    print(S)
    print(y)
    calibrator = FixedDiagonalDirichletCalibrator()
    calibrator.fit(S, y)
    print(calibrator.predict_proba(S))



import logging
from sklearn.base import BaseEstimator, RegressorMixin

import numpy as np
from .multinomial import MultinomialRegression
from ..utils import clip_for_log
from sklearn.metrics import log_loss

from .multinomial import _get_identity_weights


class TemperatureScaling(BaseEstimator, RegressorMixin):
    def __init__(self, reg_lambda_list=[0.0], reg_mu_list=[None], 
                 logit_input=False, logit_constant=None, 
                 weights_init=None, initializer='identity',
                 ref_row=True):
        self.weights_init = weights_init
        self.logit_input = logit_input
        self.logit_constant = logit_constant
        self.reg_lambda_list = reg_lambda_list
        self.reg_mu_list = reg_mu_list
        self.initializer = initializer
        self.ref_row = ref_row

    def __setup(self):
        self.calibrator_ = None
        self.reg_lambda = 0.0
        self.reg_mu = None
        self.weights_ = self.weights_init


    def fit(self, X, y, X_val=None, y_val=None, *args, **kwargs):

        self.__setup()

        k = np.shape(X)[1]

        if X_val is None:
            X_val = X.copy()
            y_val = y.copy()

        if self.logit_input == False:
            _X = np.copy(X)
            _X = np.log(clip_for_log(_X))
            _X_val = np.copy(X_val)
            _X_val = np.log(clip_for_log(X_val))
            if self.logit_constant is None:
                _X = _X - _X[:, -1].reshape(-1, 1).repeat(k, axis=1)
                _X_val = _X_val[:, -1].reshape(-1, 1).repeat(k, axis=1)
            else:
                _X = _X - self.logit_constant
                _X_val = _X_val - self.logit_constant
        else:
            _X = np.copy(X)
            _X_val = np.copy(X_val)

        for i in range(0, len(self.reg_lambda_list)):
            for j in range(0, len(self.reg_mu_list)):
                tmp_cal = MultinomialRegression(method='FixDiag', 
                                                reg_lambda=self.reg_lambda_list[i],
                                                reg_mu=self.reg_mu_list[j],
                                                ref_row=self.ref_row)
                tmp_cal.fit(_X, y, *args, **kwargs)
                tmp_loss = log_loss(y_val, tmp_cal.predict_proba(_X_val))
                
                if (i + j) == 0:
                    final_cal = tmp_cal
                    final_loss = tmp_loss
                    final_reg_lambda = self.reg_lambda_list[i]
                    final_reg_mu = self.reg_mu_list[j]
                elif tmp_loss < final_loss:
                    final_cal = tmp_cal
                    final_loss = tmp_loss
                    final_reg_lambda = self.reg_lambda_list[i]
                    final_reg_mu = self.reg_mu_list[j]

        self.calibrator_ = final_cal
        self.reg_lambda = final_reg_lambda
        self.reg_mu = final_reg_mu
        self.weights_ = self.calibrator_.weights_

        return self

    @property
    def coef_(self):
        return self.calibrator_.coef_

    @property
    def intercept_(self):
        return self.calibrator_.intercept_

    def predict_proba(self, S):
        k = np.shape(S)[1]

        if self.logit_input == False:
            _S = np.log(clip_for_log(np.copy(S)))
            if self.logit_constant is None:
                _S = _S - _S[:, -1].reshape(-1, 1).repeat(k, axis=1)
            else:
                _S = _S - self.logit_constant
        else:
            _S = np.copy(S)

        return np.asarray(self.calibrator_.predict_proba(_S))

    def predict(self, S):
        k = np.shape(S)[1]

        if self.logit_input == False:
            _S = np.log(clip_for_log(np.copy(S)))
            if self.logit_constant is None:
                _S = _S - _S[:, -1].reshape(-1, 1).repeat(k, axis=1)
            else:
                _S = _S - self.logit_constant
        else:
            _S = np.copy(S)

        return np.asarray(self.calibrator_.predict(_S))



import setuptools
from distutils.util import convert_path

with open("README.md", 'r') as f:
    long_description = f.read()

main_ns = {}
ver_path = convert_path('dirichletcal/version.py')
with open(ver_path) as ver_file:
    exec(ver_file.read(), main_ns)

setuptools.setup(
    name='dirichletcal',
    version=main_ns['__version__'],
    author='Miquel Perello Nieto and Hao Song',
    author_email='perello.nieto@gmail.com',
    description='Python code for Dirichlet calibration',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/dirichletcal/dirichlet_python',
    packages=setuptools.find_packages(),
    download_url = 'https://github.com/dirichletcal/dirichlet_python/archive/{}.tar.gz'.format(main_ns['__version__']),
    keywords = ['classifier', 'calibration', 'dirichlet', 'multiclass',
              'probability'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires = [
        'numpy>=1.14.2'
        'scipy>=1.0.0'
        'scikit-learn>=0.19.1'
        'jax'
        'jaxlib'
        'autograd'
    ]
)



import unittest
import numpy as np
from dirichletcal.calib.tempscaling import TemperatureScaling
from . import get_simple_binary_example
from . import get_extreme_binary_example
from . import get_simple_ternary_example

from sklearn.metrics import accuracy_score


class TestTemperatureScaling(unittest.TestCase):
    def setUp(self):
        self.cal = TemperatureScaling(logit_constant=0.0)

    def test_fit_predict(self):
        S, y = get_simple_binary_example()
        self.cal = TemperatureScaling(logit_constant=0.0)
        self.cal.fit(S, y)
        predictions = self.cal.predict_proba(S).argmax(axis=1)
        acc = accuracy_score(y, predictions)
        self.assertGreater(acc, 0.99, "accuracy must be superior to 99 percent")

        S, y = get_simple_ternary_example()
        self.cal = TemperatureScaling(logit_constant=0.0)
        self.cal.fit(S, y)
        predictions = self.cal.predict_proba(S).argmax(axis=1)
        acc = accuracy_score(y, predictions)
        self.assertGreater(acc, 0.97, "accuracy must be superior to 99 percent")

    def test_extreme_values(self):
        S, y = get_extreme_binary_example()
        self.cal = TemperatureScaling(logit_constant=0.0)
        self.cal.fit(S, y)
        predictions = self.cal.predict_proba(S).argmax(axis=1)
        acc = accuracy_score(y, predictions)
        self.assertGreater(acc, 0.99, "accuracy must be superior to 99 percent")


def main():
    unittest.main()


if __name__ == '__main__':
    main()
